from typing import Union
from typing import Callable
from typing import Any
from typing import Literal
import re

from laktory._logger import get_logger
from laktory.models.transformers.basechainnode import BaseChainNode
from laktory.models.transformers.basechainnode import BaseChainNodeColumn
from laktory.models.transformers.basechainnode import BaseChainNodeFuncArg
from laktory.models.transformers.basechainnode import BaseChainNodeSQLExpr
from laktory.spark import SparkColumn
from laktory.spark import SparkDataFrame


logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class SparkChainNodeFuncArg(BaseChainNodeFuncArg):
    """
    Spark function argument

    Attributes
    ----------
    value:
        Value of the argument
    """

    value: Union[Any]

    def eval(self, spark=None):
        from laktory.models.datasources.basedatasource import BaseDataSource

        v = self.value

        if isinstance(v, BaseDataSource):
            v = self.value.read(spark=spark)

        elif isinstance(v, str):

            # Imports required to evaluate expressions
            import pyspark.sql.functions as F
            from pyspark.sql.functions import lit
            from pyspark.sql.functions import col
            from pyspark.sql.functions import expr

            targets = ["lit(", "col(", "expr(", "F."]

            for f in targets:
                if f in v:
                    v = eval(v)
                    break

        return v


class SparkChainNodeColumn(BaseChainNodeColumn):
    """
    Chain node column definition

    Attributes
    ----------
    name:
        Column name
    type:
        Column data type
    unit:
        Column units
    expr:
        String representation of a polars expression
    sql_expr:
        SQL expression
    """

    name: str
    type: Union[str, None] = "string"
    unit: Union[str, None] = None
    expr: Union[str, None] = None
    sql_expr: Union[str, None] = None

    def eval(self, udfs=None):

        # Adding udfs to global variables
        if udfs is None:
            udfs = {}
        for k, v in udfs.items():
            globals()[k] = v

        # Imports required to evaluate expressions
        import pyspark.sql.functions as F
        from pyspark.sql.functions import col
        from pyspark.sql.functions import lit

        if self.sql_expr:
            return F.expr(self.sql_expr)

        expr = eval(self.expr)

        # Cleaning up global variables
        for k, v in udfs.items():
            del globals()[k]

        return expr


class SparkChainNodeSQLExpr(BaseChainNodeSQLExpr):
    """
    Chain node SQL expression

    Attributes
    ----------
    expr:
        SQL expression
    """

    def parsed_expr(self, df_id="df") -> str:
        expr = self.expr.replace("{df}", df_id)
        pattern = r"\{nodes\.(.*?)\}"
        matches = re.findall(pattern, expr)
        for m in matches:
            expr = expr.replace("{nodes." + m + "}", f"nodes__{m}")
        return expr

    def eval(self, df, chain_node=None):

        # We wanted to use parametrized queries to inject dataframes into the
        # query, but it does not seem to be mature enough to do so:
        # - as of Spark 3.5, spark connect does not support kwargs in .sql
        # - DLT does not currently support parametrized queries. Even if we
        #   bypass DLT custom sql function, the method org.apache.spark.sql.internal.CatalogImpl.dropTempView used in
        #   parametrized queries is not whitelisted and result in a py4j.security.Py4JSecurityException
        #
        # For now, we will create temp views manually.

        _spark = df.sparkSession

        # Get pipeline node if executed from pipeline
        pipeline_node = None
        if chain_node is not None:
            spark_chain = chain_node._parent
            if spark_chain is not None:
                pipeline_node = spark_chain._parent

        # Set df id (to avoid temp view with conflicting names)
        df_id = "df"
        if pipeline_node:
            df_id = f"df_{pipeline_node.name}"

        # Create views
        df.createOrReplaceTempView(df_id)
        for source in self.node_data_sources:
            _df = source.read(spark=_spark)
            _df.createOrReplaceTempView(f"nodes__{source.node.name}")

        # Run query
        return _spark.laktory.sql(self.parsed_expr(df_id))

        # class Nodes:
        #     pass
        #
        # nodes = Nodes()
        # for source in self.node_data_sources:
        #     setattr(nodes, source.node.name, source.read(spark=df.sparkSession))
        # return df.sparkSession.laktory.sql(self.parsed_expr, df=df, nodes=nodes)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class SparkChainNode(BaseChainNode):
    """
    PolarsChain node that output a dataframe upon execution. As a convenience,
    `with_column` argument can be specified to create a new column from a
    spark or sql expression. Each node is executed sequentially in the
    provided order. A node may also be another Spark Chain.

    Attributes
    ----------
    func_args:
        List of arguments to be passed to the spark function. If the function
        expects a spark column, its string representation can be provided
        with support for `col`, `lit`, `expr` and `F.`.
    func_kwargs:
        List of keyword arguments to be passed to the spark function. If the
        function expects a spark column, its string representation can be
        provided with support for `col`, `lit`, `expr` and `F.`.
    func_name:
        Name of the spark function to build the dataframe. Mutually
        exclusive to `sql_expr` and `with_column`.
    sql_expr:
        SQL Expression using `{df}` to reference upstream dataframe and
        defining how to build the output dataframe. Mutually exclusive to
        `func_name` and `with_column`. Other pipeline nodes can also be
        referenced using {nodes.node_name}.
    with_column:
        Syntactic sugar for adding a column. Mutually exclusive to `func_name`
        and `sql_expr`.
    with_columns:
        Syntactic sugar for adding columns. Mutually exclusive to `func_name`
        and `sql_expr`.

    Examples
    --------
    ```py
    import pandas as pd
    from laktory import models

    df0 = spark.createDataFrame(pd.DataFrame({"x": [1, 2, 2, 3]}))

    node = models.SparkChainNode(
        with_column={
            "name": "cosx",
            "type": "double",
            "expr": "F.cos('x')",
        },
    )
    df = node.execute(df0)

    node = models.SparkChainNode(
        with_column={
            "name": "xy",
            "type": "double",
            "expr": "F.coalesce('x')",
        },
    )
    df = node.execute(df)

    print(df.toPandas().to_string())
    '''
       x      cosx   xy
    0  1  0.540302  1.0
    1  2 -0.416147  2.0
    2  2 -0.416147  2.0
    3  3 -0.989992  3.0
    '''

    node = models.SparkChainNode(
        func_name="drop_duplicates",
        func_args=[["x"]],
    )
    df = node.execute(df)

    print(df.toPandas().to_string())
    '''
       x      cosx   xy
    0  1  0.540302  1.0
    1  2 -0.416147  2.0
    2  3 -0.989992  3.0
    '''
    ```
    """

    dataframe_type: Literal["SPARK"] = "SPARK"
    func_args: list[Union[Any]] = []
    func_kwargs: dict[str, Union[Any]] = {}
    func_name: Union[str, None] = None
    sql_expr: Union[str, None] = None
    with_column: Union[SparkChainNodeColumn, None] = None
    with_columns: Union[list[SparkChainNodeColumn], None] = []
    _parent: "SparkChain" = None
    _parsed_func_args: list = None
    _parsed_func_kwargs: dict = None
    _parsed_sql_expr: SparkChainNodeSQLExpr = None

    @property
    def parsed_func_args(self):
        if not self._parsed_func_args:
            self._parsed_func_args = [
                SparkChainNodeFuncArg(value=a) for a in self.func_args
            ]
        return self._parsed_func_args

    @property
    def parsed_func_kwargs(self):
        if not self._parsed_func_kwargs:
            self._parsed_func_kwargs = {
                k: SparkChainNodeFuncArg(value=v) for k, v in self.func_kwargs.items()
            }
        return self._parsed_func_kwargs

    @property
    def parsed_sql_expr(self):
        if self.sql_expr is None:
            return None
        if not self._parsed_sql_expr:
            self._parsed_sql_expr = SparkChainNodeSQLExpr(expr=self.sql_expr)
        return self._parsed_sql_expr

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def execute(
        self,
        df: SparkDataFrame,
        udfs: list[Callable[[...], Union[SparkColumn, SparkDataFrame]]] = None,
    ) -> Union[SparkDataFrame]:
        """
        Execute spark chain node

        Parameters
        ----------
        df:
            Input dataframe
        udfs:
            User-defined functions

        Returns
        -------
            Output dataframe
        """
        import pyspark.sql.functions as F
        from pyspark.sql.dataframe import DataFrame
        from pyspark.sql.connect.dataframe import DataFrame as DataFrameConnect
        from pyspark.sql import Column
        from laktory.spark import DATATYPES_MAP

        if udfs is None:
            udfs = []
        udfs = {f.__name__: f for f in udfs}

        # Build Columns
        if self._with_columns:
            for column in self._with_columns:
                logger.info(
                    f"Building column {column.name} as {column.expr or column.sql_expr}"
                )
                _col = column.eval(udfs=udfs)
                if column.type:
                    _col = _col.cast(DATATYPES_MAP[column.type])
                df = df.withColumns({column.name: _col})
            return df

        # From SQL expression
        if self.sql_expr:
            return self.parsed_sql_expr.eval(df, chain_node=self)

        # Get Function
        func_name = self.func_name
        if self.func_name is None:
            raise ValueError(
                "`func_name` must be specified if `sql_expr` is not specified"
            )

        # Get from UDFs
        f = udfs.get(func_name, None)

        # Get from built-in spark and spark extension (including Laktory) functions
        input_df = True
        if f is None:
            # Get function from namespace extension
            if "." in func_name:
                input_df = False
                vals = func_name.split(".")
                f = getattr(getattr(df, vals[0]), vals[1], None)
            else:
                f = getattr(type(df), func_name, None)

        if f is None:
            raise ValueError(f"Function {func_name} is not available")

        _args = self.parsed_func_args
        _kwargs = self.parsed_func_kwargs

        # Build log
        func_log = f"{func_name}("
        func_log += ",".join([a.signature() for a in _args])
        func_log += ",".join([f"{k}={a.signature()}" for k, a in _kwargs.items()])
        func_log += ")"
        logger.info(f"DataFrame {self.id} as {func_log}")

        # Build args
        args = []
        for i, _arg in enumerate(_args):
            args += [_arg.eval(spark=df.sparkSession)]

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.eval(spark=df.sparkSession)

        # Call function
        if input_df:
            df = f(df, *args, **kwargs)
        else:
            df = f(*args, **kwargs)

        return df



from uuid import UUID
import polars as pl
import numpy as np
import pytest
import laktory

df0 = pl.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
        "pi": [np.pi] * 3,
        "p": [2.0, 0.2, 0.05],
        "word": ["dog_cat", "dog_cat_mouse", "dog"],
    },
)


def atest_coalesce(df0=df0):
    return
    # TODO: re-introduce
    df = df0.with_columns(xb0=pl.expr.laktory.coalesce(pl.col("x"), pl.col("b")))
    df = df.with_columns(
        xb1=pl.expr.laktory.coalesce(pl.col("x"), pl.col("b"), available_columns=["b"])
    )
    df = df.with_columns(
        xb2=pl.expr.laktory.coalesce(["x", "b"], available_columns=["b"])
    )
    df = df.with_columns(
        xb3=pl.expr.laktory.coalesce(
            pl.col("x").sqrt(), pl.col("b").sqrt(), available_columns=["b"]
        )
    )
    assert df["xb0"].to_list() == [1, 2, 3]
    assert df["xb1"].to_list() == [2, 0, 2]
    assert df["xb2"].to_list() == [2, 0, 2]
    assert df["xb3"].to_list() == [np.sqrt(2), 0, np.sqrt(2)]


def test_compare(df0=df0):
    df = df0.with_columns(compare1=pl.expr.laktory.compare(pl.col("x"), pl.col("a")))
    df = df.with_columns(
        compare2=pl.expr.laktory.compare(
            pl.col("x"), pl.col("a"), operator=">", where=pl.col("a") > 0
        )
    )
    assert df["compare1"].to_list() == [True, False, False]
    assert df["compare2"].to_list() == [False, None, True]


def test_roundp(df0=df0):

    df = df0.with_columns(roundp_1=pl.Expr.laktory.roundp(pl.col("pi"), p=0.2))
    df = df.with_columns(roundp_11=pl.col("pi").laktory.roundp(p=0.2))
    df = df.with_columns(roundp_2=pl.Expr.laktory.roundp(pl.col("pi"), p=pl.col("p")))

    assert df["roundp_1"].to_list() == pytest.approx([3.2, 3.2, 3.2], abs=0.0001)
    assert df["roundp_11"].to_list() == pytest.approx([3.2, 3.2, 3.2], abs=0.0001)
    assert df["roundp_2"].to_list() == pytest.approx([4, 3.20, 3.15], abs=0.0001)


def test_row_number():

    df = pl.DataFrame(
        {
            "x": ["a", "a", "b", "b", "b", "c"],
            "z": ["11", "10", "22", "21", "20", "30"],
        }
    )
    df = df.with_columns(y1=pl.Expr.laktory.row_number())
    df = df.with_columns(y2=pl.Expr.laktory.row_number().over("x"))
    df = df.with_columns(y3=pl.Expr.laktory.row_number().over("x").sort_by("z"))
    assert df["y1"].to_list() == [1, 2, 3, 4, 5, 6]
    assert df["y2"].to_list() == [1, 2, 1, 2, 3, 1]
    assert df["y3"].to_list() == [2, 1, 3, 2, 1, 1]


def test_sql_expr():

    expr0 = (pl.col("data").struct.field("open") >= 2) & (pl.col("x") > 0) | (
        pl.col("symbol") == "AAPL"
    )
    expr1 = pl.Expr.laktory.sql_expr("data.open >= 2 AND x > 0 OR symbol == 'AAPL'")

    assert str(expr1) == str(expr0)


def test_string_split(df0=df0):
    df = df0.with_columns(split_1=pl.Expr.laktory.string_split(pl.col("word"), "_", 0))
    df = df.with_columns(split_2=pl.Expr.laktory.string_split(pl.col("word"), "_", 1))

    assert df["split_1"].to_list() == ["dog", "dog", "dog"]
    assert df["split_2"].to_list() == ["cat", "cat", None]


def test_uuid(df0=df0):
    df = df0.with_columns(uuid=pl.Expr.laktory.uuid())

    for _uuid in df["uuid"]:
        assert str(UUID(_uuid)) == _uuid

    assert df["uuid"].n_unique() == 3


def test_units(df0=df0):
    df = df0.with_columns(ft=pl.Expr.laktory.convert_units(pl.col("x"), "m", "ft"))
    df = df.with_columns(kelvin=pl.Expr.laktory.convert_units(pl.col("x"), "C", "K"))

    assert df["ft"].to_list() == [
        3.280839895013124,
        6.561679790026248,
        9.842519685039372,
    ]
    assert df["kelvin"].to_list() == [274.15, 275.15, 276.15]


if __name__ == "__main__":
    atest_coalesce()
    test_compare()
    test_roundp()
    test_row_number()
    test_sql_expr()
    test_string_split()
    test_uuid()
    test_units()



import re
import polars as pl


def has_column(df: pl.DataFrame, col: str) -> bool:
    """
    Check if column `col` exists in `df`

    Parameters
    ----------
    df:
        Input DataFrame
    col
        Column name

    Returns
    -------
    :
        Result

    Examples
    --------

    ```py
    import laktory  # noqa: F401
    import polars as pl

    df = pl.DataFrame(
        {
            "indexx": [1, 2, 3],
            "stock": [
                {"symbol": "AAPL", "name": "Apple"},
                {"symbol": "MSFT", "name": "Microsoft"},
                {"symbol": "GOOGL", "name": "Google"},
            ],
            "prices": [
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
            ],
        }
    )

    print(df.laktory.has_column("symbol"))
    #> False
    print(df.laktory.has_column("`stock`.`symbol`"))
    #> True
    print(df.laktory.has_column("`prices[2]`.`close`"))
    #> True
    ```
    """
    _col = re.sub(r"\[(\d+)\]", r"[*]", col)
    _col = re.sub(r"`", "", _col)
    return _col in df.laktory.schema_flat()



from typing import Any
from typing import Union
from typing import Literal
from pydantic import model_validator
from pydantic import Field
from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.grants import Grants
from laktory.models.grants.tablegrant import TableGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.resources.databricks.column import Column

logger = get_logger(__name__)


class TableLookup(ResourceLookup):
    """
    Attributes
    ----------
    name:
        Full name of the databricks_table: `catalog`.`schema`.`table`
    """

    name: str = Field(serialization_alias="id")


class Table(BaseModel, PulumiResource, TerraformResource):
    """
    A table resides in the third layer of Unity Catalog’s three-level
    namespace. It contains rows of data.

    Attributes
    ----------
    catalog_name:
        Name of the catalog storing the table
    columns:
        List of columns stored in the table
    comment:
        Text description of the catalog
    data_source_format:
        External tables are supported in multiple data source formats. The string constants identifying these formats
        are DELTA, CSV, JSON, AVRO, PARQUET, ORC, TEXT. Change forces creation of a new resource. Not supported for
        MANAGED tables or VIEW.
    grants:
        List of grants operating on the schema
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    name:
        Name of the table
    primary_key:
        Name of the column storing a unique identifier for each row. It is used
        by the builder to drop duplicated rows.
    schema_name:
        Name of the schema storing the table
    storage_credential_name:
        For EXTERNAL Tables only: the name of storage credential to use. Change forces creation of a new resource.
    storage_location:
        URL of storage location for Table data (required for EXTERNAL Tables). Not supported for VIEW or MANAGED
        table_type.
    table_type:
        Distinguishes a view vs. managed/external Table.
    view_definition:
        SQL text defining the view (for `table_type == "VIEW"`). Not supported
        for MANAGED or EXTERNAL table_type.
    warehouse_id:
        All table CRUD operations must be executed on a running cluster or SQL
        warehouse. If a warehouse_id is specified, that SQL warehouse will be
        used to execute SQL commands to manage this table.

    Examples
    --------
    ```py
    from laktory import models

    table = models.resources.databricks.Table(
        name="slv_stock_prices",
    )
    ```

    References
    ----------

    * [Databricks Unity Table](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#tables)
    * [Pulumi Databricks Table](https://www.pulumi.com/registry/packages/databricks/api-docs/sqltable/)
    """

    catalog_name: Union[str, None] = None
    columns: Union[list[Column], None] = None
    comment: Union[str, None] = None
    data_source_format: str = "DELTA"
    grants: list[TableGrant] = None
    lookup_existing: TableLookup = Field(None, exclude=True)
    name: str
    properties: Union[dict[str, str], None] = None
    primary_key: Union[str, None] = None
    schema_name: Union[str, None] = None
    storage_credential_name: Union[str, None] = None
    storage_location: Union[str, None] = None
    table_type: Literal["MANAGED", "EXTERNAL", "VIEW"] = "MANAGED"
    view_definition: Union[str, None] = None
    warehouse_id: Union[str, None] = None

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    @model_validator(mode="after")
    def assign_catalog_schema(self) -> Any:
        # Assign to columns
        if self.columns:
            for c in self.columns:
                c.table_name = self.name
                c.catalog_name = self.catalog_name
                c.schema_name = self.schema_name

        # Warehouse ID
        if self.warehouse_id is None:
            self.warehouse_id = settings.databricks_warehouse_id

        # Because properties are dynamically changed by Databricks (for Delta operations)
        # we ignore properties by default for triggering a resources change
        # Unfortunately, this did not work because ingore_changes does not apply to changes
        # made outside of terraform.
        # if "properties" not in self.options.model_fields_set:
        #     # self.options.ignore_changes = ["properties"]
        #     self.properties = {"a": "b"}

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""

        if self.lookup_existing:
            return self.lookup_existing.name

        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def column_names(self) -> list[str]:
        """List of column names"""
        return [c.name for c in self.columns]

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        return self.builder.is_from_cdc

    # ----------------------------------------------------------------------- #
    #  Methods                                                                #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.table)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - table grants
        """
        resources = []

        # Schema grants
        if self.grants:
            resources += [
                Grants(
                    resource_name=f"grants-{self.resource_name}",
                    table=self.full_name,
                    grants=[
                        {"principal": g.principal, "privileges": g.privileges}
                        for g in self.grants
                    ],
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlTable"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "columns",
            "grants",
            "primary_key",
        ]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        if not self.columns:
            return d
        d["columns"] = []
        for i, c in enumerate(self.columns):
            d["columns"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_table"

    @property
    def terraform_resource_lookup_type(self) -> str:
        return "databricks_table"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        if not self.columns:
            return d
        d["column"] = []
        for i, c in enumerate(self.columns):
            d["column"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d



from .basemodel import BaseModel
from .dataevent import DataEvent
from .dataproducer import DataProducer
from .datasinks import *
from .datasources import *
from .grants import *
from .pipeline import Pipeline
from .pipelinenode import PipelineNode
from .pipelinenodeexpectation import PipelineNodeExpectation
from .resources import *
from .transformers import *
from .stacks import *



# COMMAND ----------
dbutils.widgets.text("pipeline_name", "pl-stock-prices")
dbutils.widgets.text("node_name", "")

# COMMAND ----------
import importlib
import sys
import os
import pyspark.sql.functions as F

from laktory import models
from laktory import get_logger
from laktory import settings

logger = get_logger(__name__)

# --------------------------------------------------------------------------- #
# Read Pipeline                                                               #
# --------------------------------------------------------------------------- #

pl_name = dbutils.widgets.get("pipeline_name")
node_name = dbutils.widgets.get("node_name")
filepath = f"/Workspace{settings.workspace_laktory_root}pipelines/{pl_name}.json"
with open(filepath, "r") as fp:
    pl = models.Pipeline.model_validate_json(fp.read())


# Import User Defined Functions
sys.path.append(f"/Workspace{settings.workspace_laktory_root}pipelines/")
udfs = []
for udf in pl.udfs:
    if udf.module_path:
        sys.path.append(os.path.abspath(udf.module_path))
    module = importlib.import_module(udf.module_name)
    module = importlib.reload(module)
    udfs += [getattr(module, udf.function_name)]

# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

if node_name:
    pl.nodes_dict[node_name].execute(spark=spark, udfs=udfs)
else:
    pl.execute(spark=spark, udfs=udfs)



import laktory
from laktory.models import BaseModel
from laktory._parsers import remove_empty
from laktory._parsers import merge_dicts


class Camel(BaseModel):
    d: dict = {}
    l: list = []


def test_camel_case():
    camel = Camel(
        d={"this_is_a_test": ["value_a", "value_b", {"key_alpha": 0, "keyBeta": 1}]},
        l=["a", "a-b-c", "class_member"],
    )
    d = laktory._parsers.camelize_keys(camel.model_dump())
    print(d)
    assert d == {
        "d": {"thisIsATest": ["value_a", "value_b", {"keyBeta": 1, "keyAlpha": 0}]},
        "l": ["a", "a-b-c", "class_member"],
    }


def test_remove_empty():
    d = {
        "a": "a",
        "b": {},
        "c": [{}, "c", []],
        "d": [],
    }

    assert remove_empty(d) == {"a": "a", "c": ["c"]}


def test_merge_dicts():

    baseline = {
        "name": "Apple",
        "symbols": {
            "nasdaq": "AAPL",
            "sp500": "APPL",
        },
        "prices": [
            {
                "timestamp": 0,
                "open": 1,
                "close": 2,
            },
            {
                "timestamp": 1,
                "open": 3,
                "close": 4,
            },
            {
                "timestamp": 2,
                "open": 5,
                "close": 6,
            },
        ],
    }

    overwrite = {
        "symbols": {
            "sp500": "AAPL",
        },
        "prices": {
            1: {
                "close": 5,
            }
        },
    }

    result = merge_dicts(baseline, overwrite)

    assert result == {
        "name": "Apple",
        "symbols": {"nasdaq": "AAPL", "sp500": "AAPL"},
        "prices": [
            {"timestamp": 0, "open": 1, "close": 2},
            {"timestamp": 1, "open": 3, "close": 5},
            {"timestamp": 2, "open": 5, "close": 6},
        ],
    }


if __name__ == "__main__":
    test_camel_case()
    test_remove_empty()
    test_merge_dicts()



from typing import Literal
from laktory.models.basemodel import BaseModel


class RegisteredModelGrant(BaseModel):
    """
    Privileges granted to a principal and operating on a registered model

    Attributes
    ----------
    principal
        User, group or service principal name
    privileges
        List of allowed privileges

    References
    ----------
    * [privilege types](https://docs.databricks.com/en/sql/language-manual/sql-ref-privileges.html#privilege-types)
    """

    principal: str
    privileges: list[Literal["ALL_PRIVILEGES", "EXECUTE"]]



import polars as pl
from datetime import datetime

df = pl.DataFrame(
    {
        "i": [1, 2, 3],
        "d": [
            datetime(2025, 1, 1),
            datetime(2025, 1, 2),
            datetime(2025, 1, 3),
        ],
        "f": [4.0, 5.0, 6.0],
        "s": ["a", "b", "c"],
    }
)

print(df)



import datetime
import os
import polars as pl

import laktory
from laktory._testing.stockprices import df_slv_polars
from laktory._testing.stockprices import df_meta_polars

df = pl.DataFrame(
    [
        (
            1,
            ["a", "b"],
            {"id": 3, "email": "@gmail.com"},
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        ),
        (
            2,
            ["b", "c"],
            {"id": 2, "email": "@gmail.com"},
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        ),
        (
            3,
            ["c", "d"],
            {"id": 1, "email": "@gmail.com"},
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
            [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        ),
    ],
    schema=["x@x", "y", "z", "u", "u2"],
)


def test_df_schema_flat():

    schema = df.laktory.schema_flat()
    assert schema == [
        "x@x",
        "y",
        "z",
        "z.id",
        "z.email",
        "u",
        "u[*].a",
        "u[*].b",
        "u2",
        "u2[*].a",
        "u2[*].b",
    ]


def test_df_has_column():
    assert df.laktory.has_column("x@x")
    assert df.laktory.has_column("`x@x`")
    assert df.laktory.has_column("u[0].a")


def test_union():
    df2 = df.laktory.union(df)

    assert df2.height == df.height * 2
    assert df2.schema == df.schema


def test_join():
    left = df_slv_polars.filter(
        pl.col("created_at") == datetime.datetime(2023, 9, 1)
    ).filter(pl.col("symbol") != "GOOGL")
    other = df_meta_polars.rename({"symbol2": "symbol"})

    df = left.laktory.smart_join(
        other=other,
        on=["symbol"],
        how="left",
    )

    # Test join columns uniqueness
    _df = df.with_columns(symbol2=pl.lit("a"))
    _df = df.with_columns(symbol=pl.lit("a"))
    _df = _df.select("symbol")

    # Test data
    data = df.to_dict(as_series=False)
    print(data)
    assert data == {
        "created_at": [
            datetime.datetime(2023, 9, 1, 0, 0),
            datetime.datetime(2023, 9, 1, 0, 0),
            datetime.datetime(2023, 9, 1, 0, 0),
        ],
        "symbol": ["AAPL", "AMZN", "MSFT"],
        "open": [189.49000549316406, 139.4600067138672, 331.30999755859375],
        "close": [189.4600067138672, 138.1199951171875, 328.6600036621094],
        "currency": ["USD", "USD", None],
        "first_traded": ["1980-12-12T14:30:00.000Z", "1997-05-15T13:30:00.000Z", None],
    }

    # Left and Other on
    df2 = left.laktory.smart_join(
        other=df_meta_polars,
        left_on="symbol",
        other_on="symbol2",
    )
    assert "symbol" in df2.columns
    assert "symbol2" not in df2.columns
    assert df2.select(df.columns).equals(df)

    # Not Coalesce
    df4 = left.with_columns(open=pl.lit(None)).laktory.smart_join(
        other=df_meta_polars.with_columns(open=pl.lit(2)),
        left_on="symbol",
        other_on="symbol2",
    )
    assert df4.columns == [
        "created_at",
        "symbol",
        "open",
        "close",
        "currency",
        "first_traded",
        "open_right",
    ]

    # With Coalesce
    df5 = left.with_columns(open=pl.lit(None)).laktory.smart_join(
        other=df_meta_polars.with_columns(open=pl.lit(2)),
        left_on="symbol",
        other_on="symbol2",
        coalesce=True,
    )
    assert df5.columns == [
        "created_at",
        "symbol",
        "open",
        "close",
        "currency",
        "first_traded",
    ]
    assert df5["open"].fill_null(-1).to_list() == [2, 2, -1]


def test_join_outer():
    left = df_slv_polars.filter(
        pl.col("created_at") == datetime.datetime(2023, 9, 1)
    ).filter(pl.col("symbol") != "GOOGL")
    other = df_meta_polars.rename({"symbol2": "symbol"})

    df = left.laktory.smart_join(
        other=other,
        on=["symbol"],
        how="full",
    )

    # Test join columns uniqueness
    _df = df.with_columns(symbol2=pl.lit("a"))
    _df = df.with_columns(symbol=pl.lit("a"))
    _df = _df.select("symbol")

    # Test data
    data = df.to_dict(as_series=False)
    print(data)
    assert data == {
        "created_at": [
            datetime.datetime(2023, 9, 1, 0, 0),
            None,
            datetime.datetime(2023, 9, 1, 0, 0),
            datetime.datetime(2023, 9, 1, 0, 0),
        ],
        "symbol": ["AAPL", "GOOGL", "AMZN", "MSFT"],
        "open": [189.49000549316406, None, 139.4600067138672, 331.30999755859375],
        "close": [189.4600067138672, None, 138.1199951171875, 328.6600036621094],
        "currency": ["USD", "USD", "USD", None],
        "first_traded": [
            "1980-12-12T14:30:00.000Z",
            "2004-08-19T13:30:00.00Z",
            "1997-05-15T13:30:00.000Z",
            None,
        ],
    }


def test_aggregation():
    _df = df_slv_polars.filter(pl.col("created_at") < datetime.datetime(2023, 9, 7))

    # Symbol
    df = _df.laktory.groupby_and_agg(
        groupby_columns=["symbol"],
        agg_expressions=[
            {
                "name": "mean_close",
                "expr": "pl.col('close').mean()",
            },
        ],
    ).sort("symbol")
    assert "symbol" in df.columns
    assert df["mean_close"].round(2).to_list() == [187.36, 136.92, 135.3, 331.7]


def test_window_filter():
    df = df_slv_polars.laktory.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=2,
    ).select("created_at", "symbol", "_row_index")

    data = df.to_dict(as_series=False)
    print(data)
    assert data == {
        "created_at": [
            datetime.datetime(2023, 9, 28, 0, 0),
            datetime.datetime(2023, 9, 29, 0, 0),
            datetime.datetime(2023, 9, 28, 0, 0),
            datetime.datetime(2023, 9, 29, 0, 0),
            datetime.datetime(2023, 9, 28, 0, 0),
            datetime.datetime(2023, 9, 29, 0, 0),
            datetime.datetime(2023, 9, 28, 0, 0),
            datetime.datetime(2023, 9, 29, 0, 0),
        ],
        "symbol": ["AAPL", "AAPL", "AMZN", "AMZN", "GOOGL", "GOOGL", "MSFT", "MSFT"],
        "_row_index": [2, 1, 2, 1, 2, 1, 2, 1],
    }


if __name__ == "__main__":
    test_df_schema_flat()
    test_df_has_column()
    test_union()
    test_join()
    test_join_outer()
    test_aggregation()
    test_window_filter()



from typing import Union
from typing import Any
from typing import Literal
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._parsers import merge_dicts
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline import Pipeline
from laktory.models.resources.baseresource import ResourceOptions
from laktory.models.resources.databricks.catalog import Catalog
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.dbfsfile import DbfsFile
from laktory.models.resources.databricks.directory import Directory
from laktory.models.resources.databricks.dltpipeline import DLTPipeline
from laktory.models.resources.databricks.externallocation import ExternalLocation
from laktory.models.resources.databricks.group import Group
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.metastore import Metastore
from laktory.models.resources.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.resources.databricks.mwsnetworkconnectivityconfig import (
    MwsNetworkConnectivityConfig,
)
from laktory.models.resources.databricks.notebook import Notebook
from laktory.models.resources.databricks.schema import Schema
from laktory.models.resources.databricks.secret import Secret
from laktory.models.resources.databricks.secretscope import SecretScope
from laktory.models.resources.databricks.serviceprincipal import ServicePrincipal
from laktory.models.resources.databricks.sqlquery import SqlQuery
from laktory.models.resources.databricks.table import Table
from laktory.models.resources.databricks.user import User
from laktory.models.resources.databricks.volume import Volume
from laktory.models.resources.databricks.warehouse import Warehouse
from laktory.models.resources.databricks.workspacefile import WorkspaceFile
from laktory.models.resources.providers.awsprovider import AWSProvider
from laktory.models.resources.providers.azureprovider import AzureProvider
from laktory.models.resources.providers.azurepulumiprovider import AzurePulumiProvider
from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.providers.databricksprovider import DatabricksProvider

logger = get_logger(__name__)

DIRPATH = "./"


class Terraform(BaseModel):
    backend: dict[str, Any] = None


class Pulumi(BaseModel):
    """
    config:
        Pulumi configuration settings. Generally used to
        configure providers. See references for more details.
    outputs:
        Requested resources-related outputs. See references for details.

    References
    ----------
    - Pulumi [configuration](https://www.pulumi.com/docs/concepts/config/)
    - Pulumi [outputs](https://www.pulumi.com/docs/concepts/inputs-outputs/#outputs)
    """

    config: dict[str, str] = {}
    outputs: dict[str, str] = {}


class StackResources(BaseModel):
    """
    Resources definition for a given stack or stack environment.

    Attributes
    ----------
    databricks_dbfsfiles:
        Databricks DbfsFiles
    databricks_catalogs:
        Databricks Catalogs
    databricks_clusters:
        Databricks Clusters
    databricks_directories:
        Databricks Directories
    databricks_externallocations:
        Databricks External Locations
    databricks_groups:
        Databricks Groups
    databricks_jobs:
        Databricks Jobs
    databricks_metastores:
        Databricks Metastores
    databricks_networkconnectivityconfig
        Databricks Network Connectivity Config
    databricks_notebooks:
        Databricks Notebooks
    databricks_dltpipelines:
        Databricks DLT Pipelines
    databricks_schemas:
        Databricks Schemas
    databricks_secretscopes:
        Databricks SecretScopes
    databricks_serviceprincipals:
        Databricks ServicePrincipals
    databricks_sqlqueries:
        Databricks SQLQueries
    databricks_tables:
        Databricks Tables
    providers:
        Providers
    databricks_users:
        Databricks Users
    databricks_volumes:
        Databricks Volumes
    databricks_warehouses:
        Databricks Warehouses
    databricks_workspacefiles:
        Databricks WorkspacFiles
    pipelines:
        Laktory Pipelines
    """

    databricks_dbfsfiles: dict[str, DbfsFile] = {}
    databricks_catalogs: dict[str, Catalog] = {}
    databricks_clusters: dict[str, Cluster] = {}
    databricks_directories: dict[str, Directory] = {}
    databricks_externallocations: dict[str, ExternalLocation] = {}
    databricks_groups: dict[str, Group] = {}
    databricks_jobs: dict[str, Job] = {}
    databricks_metastoredataaccesses: dict[str, MetastoreDataAccess] = {}
    databricks_metastores: dict[str, Metastore] = {}
    databricks_networkconnectivityconfig: dict[str, MwsNetworkConnectivityConfig] = {}
    databricks_notebooks: dict[str, Notebook] = {}
    databricks_dltpipelines: dict[str, DLTPipeline] = {}
    databricks_schemas: dict[str, Schema] = {}
    databricks_secrets: dict[str, Secret] = {}
    databricks_secretscopes: dict[str, SecretScope] = {}
    databricks_serviceprincipals: dict[str, ServicePrincipal] = {}
    databricks_sqlqueries: dict[str, SqlQuery] = {}
    databricks_tables: dict[str, Table] = {}
    databricks_users: dict[str, User] = {}
    databricks_volumes: dict[str, Volume] = {}
    databricks_warehouses: dict[str, Warehouse] = {}
    databricks_workspacefiles: dict[str, WorkspaceFile] = {}
    pipelines: dict[str, Pipeline] = {}
    providers: dict[
        str, Union[AWSProvider, AzureProvider, AzurePulumiProvider, DatabricksProvider]
    ] = {}

    @model_validator(mode="after")
    def update_resource_names(self) -> Any:
        for k, r in self._get_all().items():
            if r.resource_name_ and k != r.resource_name_:
                raise ValueError(
                    f"Provided resource name {r.resource_name_} does not match provided key {k}"
                )
            r.resource_name_ = k
        return self

    def _get_all(self, providers_excluded=False, providers_only=False):
        resources = {}
        for resource_type in self.model_fields.keys():
            if resource_type in ["variables"]:
                continue

            for resource_name, _r in getattr(self, resource_type).items():
                if providers_excluded and isinstance(_r, BaseProvider):
                    continue

                if providers_only and not isinstance(_r, BaseProvider):
                    continue

                resources[resource_name] = _r

        return resources


class EnvironmentStack(BaseModel):
    """
    Environment-specific stack definition.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.
    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    name: str
    organization: str = None
    pulumi: Pulumi = Pulumi()
    resources: Union[StackResources, None] = StackResources()
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}


class EnvironmentSettings(BaseModel):
    """
    Settings overwrite for a specific environments

    Attributes
    ----------
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    variables:
        Dictionary of variables made available in the resources definition.
    """

    resources: Any = None
    variables: dict[str, Any] = None


class Stack(BaseModel):
    """
    The Stack defines a collection of deployable resources, the deployment
    configuration, some variables and the environment-specific settings.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    environments:
        Environment-specific overwrite of config, resources or variables
        arguments.
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.

    Examples
    --------
    ```py
    from laktory import models

    stack = models.Stack(
        name="workspace",
        backend="pulumi",
        pulumi={
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
        },
        resources={
            "databricks_dltpipelines": {
                "pl-stock-prices": {
                    "name": "pl-stock-prices",
                    "development": "${vars.is_dev}",
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}},
                    ],
                }
            },
            "databricks_jobs": {
                "job-stock-prices": {
                    "name": "job-stock-prices",
                    "clusters": [
                        {
                            "name": "main",
                            "spark_version": "14.0.x-scala2.12",
                            "node_type_id": "Standard_DS3_v2",
                        }
                    ],
                    "tasks": [
                        {
                            "task_key": "ingest",
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/.laktory/jobs/ingest_stock_prices.py",
                            },
                        },
                        {
                            "task_key": "pipeline",
                            "depends_ons": [{"task_key": "ingest"}],
                            "pipeline_task": {
                                "pipeline_id": "${resources.dlt-pl-stock-prices.id}",
                            },
                        },
                    ],
                }
            },
        },
        variables={
            "org": "okube",
        },
        environments={
            "dev": {
                "variables": {
                    "is_dev": True,
                }
            },
            "prod": {
                "variables": {
                    "is_dev": False,
                }
            },
        },
    )

    print(stack)
    '''
    variables={'org': 'okube'} backend='pulumi' description=None environments={'dev': EnvironmentSettings(variables={'is_dev': True}, resources=None), 'prod': EnvironmentSettings(variables={'is_dev': False}, resources=None)} name='workspace' organization=None pulumi=Pulumi(variables={}, config={'databricks:host': '${vars.DATABRICKS_HOST}', 'databricks:token': '${vars.DATABRICKS_TOKEN}'}, outputs={}) resources=StackResources(variables={}, databricks_dbfsfiles={}, databricks_catalogs={}, databricks_clusters={}, databricks_directories={}, databricks_externallocations={}, databricks_groups={}, databricks_jobs={'job-stock-prices': Job(resource_name_='job-stock-prices', options=ResourceOptions(variables={}, depends_on=[], provider=None, ignore_changes=None, aliases=None, delete_before_replace=True, import_=None, parent=None, replace_on_changes=None), lookup_existing=None, variables={}, access_controls=[], clusters=[JobCluster(resource_name_=None, options=ResourceOptions(variables={}, depends_on=[], provider=None, ignore_changes=None, aliases=None, delete_before_replace=True, import_=None, parent=None, replace_on_changes=None), lookup_existing=None, variables={}, access_controls=None, apply_policy_default_values=None, autoscale=None, autotermination_minutes=None, cluster_id=None, custom_tags=None, data_security_mode='USER_ISOLATION', driver_instance_pool_id=None, driver_node_type_id=None, enable_elastic_disk=None, enable_local_disk_encryption=None, idempotency_token=None, init_scripts=[], instance_pool_id=None, is_pinned=None, libraries=None, name='main', node_type_id='Standard_DS3_v2', num_workers=None, policy_id=None, runtime_engine=None, single_user_name=None, spark_conf={}, spark_env_vars={}, spark_version='14.0.x-scala2.12', ssh_public_keys=[])], continuous=None, control_run_state=None, email_notifications=None, format=None, health=None, max_concurrent_runs=None, max_retries=None, min_retry_interval_millis=None, name='job-stock-prices', notification_settings=None, parameters=[], retry_on_timeout=None, run_as=None, schedule=None, tags={}, tasks=[JobTask(variables={}, condition_task=None, depends_ons=None, description=None, email_notifications=None, existing_cluster_id=None, health=None, job_cluster_key='main', libraries=None, max_retries=None, min_retry_interval_millis=None, notebook_task=JobTaskNotebookTask(variables={}, notebook_path='/.laktory/jobs/ingest_stock_prices.py', base_parameters=None, source=None), notification_settings=None, pipeline_task=None, retry_on_timeout=None, run_if=None, run_job_task=None, sql_task=None, task_key='ingest', timeout_seconds=None), JobTask(variables={}, condition_task=None, depends_ons=[JobTaskDependsOn(variables={}, task_key='ingest', outcome=None)], description=None, email_notifications=None, existing_cluster_id=None, health=None, job_cluster_key=None, libraries=None, max_retries=None, min_retry_interval_millis=None, notebook_task=None, notification_settings=None, pipeline_task=JobTaskPipelineTask(variables={}, pipeline_id='${resources.dlt-pl-stock-prices.id}', full_refresh=None), retry_on_timeout=None, run_if=None, run_job_task=None, sql_task=None, task_key='pipeline', timeout_seconds=None)], timeout_seconds=None, trigger=None, webhook_notifications=None)}, databricks_metastoredataaccesses={}, databricks_metastores={}, databricks_networkconnectivityconfig={}, databricks_notebooks={}, databricks_dltpipelines={'pl-stock-prices': DLTPipeline(resource_name_='pl-stock-prices', options=ResourceOptions(variables={}, depends_on=[], provider=None, ignore_changes=None, aliases=None, delete_before_replace=True, import_=None, parent=None, replace_on_changes=None), lookup_existing=None, variables={}, access_controls=[], allow_duplicate_names=None, catalog=None, channel='PREVIEW', clusters=[], configuration={}, continuous=None, development='${vars.is_dev}', edition=None, libraries=[PipelineLibrary(variables={}, file=None, notebook=PipelineLibraryNotebook(variables={}, path='/pipelines/dlt_brz_template.py'))], name='pl-stock-prices', notifications=[], photon=None, serverless=None, storage=None, target=None)}, databricks_schemas={}, databricks_secrets={}, databricks_secretscopes={}, databricks_serviceprincipals={}, databricks_sqlqueries={}, databricks_tables={}, databricks_users={}, databricks_volumes={}, databricks_warehouses={}, databricks_workspacefiles={}, pipelines={}, providers={}) terraform=Terraform(variables={}, backend=None)
    '''
    ```

    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    environments: dict[str, EnvironmentSettings] = {}
    name: str
    organization: Union[str, None] = None
    pulumi: Pulumi = Pulumi()
    resources: Union[StackResources, None] = StackResources()
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}
    _envs: dict[str, EnvironmentStack] = None

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def get_env(self, env_name: str, inject_vars=True) -> EnvironmentStack:
        """
        Complete definition the stack for a given environment. It takes into
        account both the default stack values and environment-specific
        overwrites.

        Parameters
        ----------
        env_name:
            Name of the environment

        Returns
        -------
        :
            Environment definitions.
        """

        if self._envs is None:

            ENV_FIELDS = ["pulumi", "resources", "terraform", "variables"]

            # Because options is an excluded field for all resources and
            # sub-resources, we need to manually dump it and add it to
            # the base dump
            def dump_with_options(obj: Any) -> Any:

                # Check data type, call recursively if not a BaseModel
                if isinstance(obj, list):
                    return [dump_with_options(v) for v in obj]
                elif isinstance(obj, dict):
                    return {k: dump_with_options(v) for k, v in obj.items()}
                elif not isinstance(obj, BaseModel):
                    return obj

                # Get model dump
                model = obj
                data = model.model_dump(exclude_unset=True)

                # Loop through all model fields
                for field_name, field in model.model_fields.items():

                    # Explicitly dump options if found in the model
                    if field_name == "options" and field.annotation == ResourceOptions:
                        data["options"] = model.options.model_dump(exclude_unset=True)

                    if field_name == "resource_name_" and model.resource_name_:
                        data["resource_name_"] = model.resource_name_

                    if field_name == "lookup_existing" and model.lookup_existing:
                        data["lookup_existing"] = model.lookup_existing.model_dump(
                            exclude_unset=True
                        )

                    # Parse list
                    if isinstance(data.get(field_name, None), list):
                        data[field_name] = [
                            dump_with_options(v) for v in getattr(model, field_name)
                        ]

                    # Parse dict (might result from a dict or a BaseModel)
                    elif isinstance(data.get(field_name, None), dict):
                        a = getattr(model, field_name)

                        if isinstance(a, dict):
                            for k in a.keys():
                                data[field_name][k] = dump_with_options(a[k])
                        else:
                            data[field_name] = dump_with_options(a)

                return data

            d = dump_with_options(self)
            _envs = d.pop("environments")

            envs = {}
            for _env_name, env in self.environments.items():
                for k in ENV_FIELDS:
                    v1 = _envs[_env_name].get(k, {})
                    if k in d:
                        d[k] = merge_dicts(d[k], v1)
                    elif k in _envs[_env_name]:
                        d[k] = v1

                # Inject Variables
                if inject_vars:
                    d = self.environments[_env_name].inject_vars(d)

                envs[_env_name] = EnvironmentStack(**d)

            self._envs = envs

        return self._envs[env_name]

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    def to_pulumi(self, env_name: Union[str, None] = None):
        """
        Create a pulumi stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : PulumiStack
            Pulumi-specific stack definition
        """
        from laktory.models.stacks.pulumistack import PulumiStack

        if env_name is not None and env_name in self.environments.keys():
            env = self.get_env(env_name=env_name, inject_vars=False)
        else:
            env = self

        # Resources
        resources = {}
        for r in env.resources._get_all().values():
            r.variables = env.variables
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        return PulumiStack(
            name=env.name,
            organization=env.organization,
            config=env.pulumi.config,
            description=env.description,
            resources=resources,
            variables=env.variables,
            outputs=env.pulumi.outputs,
        )

    # ----------------------------------------------------------------------- #
    # Terraform Methods                                                       #
    # ----------------------------------------------------------------------- #

    def to_terraform(self, env_name: Union[str, None] = None):
        """
        Create a terraform stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : TerraformStack
            Terraform-specific stack definition
        """
        from laktory.models.stacks.terraformstack import TerraformStack

        if env_name is not None and env_name in self.environments.keys():
            env = self.get_env(env_name=env_name, inject_vars=False)
        else:
            env = self

        # Providers
        providers = {}
        for r in env.resources._get_all(providers_only=True).values():
            for _r in r.core_resources:
                rname = _r.resource_name
                providers[rname] = _r

        # Resources
        resources = {}
        for r in env.resources._get_all(providers_excluded=True).values():
            r.variables = env.variables
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        # Update terraform
        return TerraformStack(
            terraform={"backend": env.terraform.backend},
            providers=providers,
            resources=resources,
            variables=env.variables,
        )



from typing import Union
from typing import Any
from typing import Literal
from pydantic import Field
from laktory.models.basemodel import BaseModel
from laktory.spark import is_spark_dataframe
from laktory.spark import SparkDataFrame
from laktory.polars import is_polars_dataframe
from laktory.polars import PolarsDataFrame
from laktory.types import AnyDataFrame


class BaseDataSink(BaseModel):
    """
    Base class for building data sink

    Attributes
    ----------
    mode:
        Write mode.
        - overwrite: Overwrite existing data
        - append: Append contents of the dataframe to existing data
        - error: Throw and exception if data already exists
        - ignore: Silently ignore this operation if data already exists
        - complete: Overwrite for streaming dataframes
    write_options:
        Other options passed to `spark.write.options`
    """

    mode: Union[Literal["OVERWRITE", "APPEND", "IGNORE", "ERROR", "COMPLETE"], None] = (
        None
    )
    write_options: dict[str, str] = {}
    _parent: "PipelineNode" = None

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self)

    # ----------------------------------------------------------------------- #
    # Writers                                                                 #
    # ----------------------------------------------------------------------- #

    def write(self, df: AnyDataFrame, mode=None) -> None:
        """
        Write dataframe into sink.

        Parameters
        ----------
        df:
            Input dataframe
        mode:
            Write mode overwrite of the sink default mode.

        Returns
        -------
        """
        if mode is None:
            mode = self.mode
        if is_spark_dataframe(df):
            self._write_spark(df, mode=mode)
        elif is_polars_dataframe(df):
            self._write_polars(df, mode=mode)
        else:
            raise ValueError()

    def _write_spark(self, df: SparkDataFrame, mode=mode) -> None:
        raise NotImplementedError("Not implemented for Spark dataframe")

    def _write_polars(self, df: PolarsDataFrame, mode=mode) -> None:
        raise NotImplementedError("Not implemented for Polars dataframe")

    # ----------------------------------------------------------------------- #
    # Sources                                                                 #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream=None):
        raise NotImplementedError()

    def read(self, spark=None, as_stream=None):
        """
        Read dataframe from sink.

        Parameters
        ----------
        spark:
            Spark Session
        as_stream:
            If `True`, dataframe read as stream.

        Returns
        -------
        """
        return self.as_source(as_stream=as_stream).read(spark=spark)



from laktory.constants import SUPPORTED_DATATYPES
from laktory.spark import DATATYPES_MAP as SPARK_DATATYPES_MAP
from laktory.polars import DATATYPES_MAP as POLARS_DATATYPES_MAP


def test_spark_datatypes():

    for dtype in SUPPORTED_DATATYPES:
        assert dtype in SPARK_DATATYPES_MAP

    for k, v in SPARK_DATATYPES_MAP.items():
        assert k in SUPPORTED_DATATYPES
        assert v is not None


def test_polars_datatypes():

    for dtype in SUPPORTED_DATATYPES:
        assert dtype in POLARS_DATATYPES_MAP

    for k, v in POLARS_DATATYPES_MAP.items():
        assert k in SUPPORTED_DATATYPES
        assert v is not None


if __name__ == "__main__":
    test_spark_datatypes()
    test_polars_datatypes()



from typing import Any
from typing import Literal
from typing import Union
from pydantic import model_validator
from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.constants import DEFAULT_DFTYPE
from laktory.spark import SparkDataFrame
from laktory.spark import is_spark_dataframe
from laktory.polars import PolarsDataFrame
from laktory.polars import is_polars_dataframe
from laktory.types import AnyDataFrame
from laktory._logger import get_logger

logger = get_logger(__name__)


class DataFrameSample(BaseModel):
    fraction: float
    seed: Union[int, None] = None


class Watermark(BaseModel):
    """
    Definition of a spark structured streaming watermark for joining data
    streams.

    Attributes
    ----------
    column:
        Event time column name
    threshold:
        How late, expressed in seconds, the data is expected to be with
        respect to event time.

    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str
    threshold: str


class DataSourceCDC(BaseModel):
    """
    Defines the change data capture (CDC) properties of a data source. They are
    used to build the target using `apply_changes` method from Databricks DLT.

    Attributes
    ----------
    apply_as_deletes:
        Specifies when a CDC event should be treated as a DELETE rather than
        an upsert. To handle out-of-order data, the deleted row is temporarily
        retained as a tombstone in the underlying Delta table, and a view is
        created in the metastore that filters out these tombstones.
    apply_as_truncates:
        Specifies when a CDC event should be treated as a full table TRUNCATE.
        Because this clause triggers a full truncate of the target table, it
        should be used only for specific use cases requiring this
        functionality.
    columns:
        A subset of columns to include in the target table. Use `columns` to
        specify the complete list of columns to include.
    except_columns:
        A subset of columns to exclude in the target table.
    ignore_null_updates:
        Allow ingesting updates containing a subset of the target columns.
        When a CDC event matches an existing row and ignore_null_updates is
        `True`, columns with a null will retain their existing values in the
        target. This also applies to nested columns with a value of null. When
        ignore_null_updates is `False`, existing values will be overwritten
        with null values.
    primary_keys:
        The column or combination of columns that uniquely identify a row in
        the source data. This is used to identify which CDC events apply to
        specific records in the target table.
    scd_type:
        Whether to store records as SCD type 1 or SCD type 2.
    sequence_by:
        The column name specifying the logical order of CDC events in the
        source data. Delta Live Tables uses this sequencing to handle change
        events that arrive out of order.
    track_history_columns:
        A subset of output columns to be tracked for history in the target table.
    track_history_except_columns:
        A subset of output columns to be excluded from tracking.

    References
    ----------
    https://docs.databricks.com/en/delta-live-tables/python-ref.html#change-data-capture-with-python-in-delta-live-tables
    """

    apply_as_deletes: Union[str, None] = None
    apply_as_truncates: Union[str, None] = None
    columns: Union[list[str], None] = []
    except_columns: Union[list[str], None] = []
    ignore_null_updates: Union[bool, None] = None
    primary_keys: list[str]
    scd_type: Literal[1, 2] = None
    sequence_by: str
    track_history_columns: Union[list[str], None] = None
    track_history_except_columns: Union[list[str], None] = None


class BaseDataSource(BaseModel):
    """
    Base class for building data source

    Attributes
    ----------
    as_stream:
        If `True`source is read as a streaming DataFrame.
    broadcast:
        If `True` DataFrame is broadcasted
    cdc:
        Change data capture specifications
    dataframe_type:
        Type of dataframe
    drops:
        List of columns to drop
    filter:
        SQL expression used to select specific rows from the source table
    renames:
        Mapping between the source table column names and new column names
    selects:
        Columns to select from the source table. Can be specified as a list
        or as a dictionary to rename the source columns
    watermark
        Spark structured streaming watermark specifications
    """

    as_stream: bool = False
    broadcast: Union[bool, None] = False
    cdc: Union[DataSourceCDC, None] = None
    dataframe_type: Literal["SPARK", "POLARS"] = DEFAULT_DFTYPE
    drops: Union[list, None] = None
    filter: Union[str, None] = None
    limit: Union[int, None] = None
    mock_df: Any = Field(default=None, exclude=True)
    renames: Union[dict[str, str], None] = None
    sample: Union[DataFrameSample, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    watermark: Union[Watermark, None] = None
    _parent: "PipelineNode" = None

    @model_validator(mode="after")
    def options(self) -> Any:

        # Overwrite Dataframe type if mock dataframe is provided
        if is_spark_dataframe(self.mock_df):
            self.dataframe_type = "SPARK"
        elif is_polars_dataframe(self.mock_df):
            self.dataframe_type = "POLARS"

        if self.dataframe_type == "SPARK":
            pass
        elif self.dataframe_type == "POLARS":
            if self.as_stream:
                raise ValueError("Polars DataFrames don't support streaming read.")
            if self.watermark:
                raise ValueError("Polars DataFrames don't support watermarking.")
            if self.broadcast:
                raise ValueError("Polars DataFrames don't support broadcasting.")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def user_dftype(self):
        if "dataframe_type" in self.model_fields_set:
            return self.dataframe_type
        return None

    @property
    def _id(self):
        return str(self)

    @property
    def is_cdc(self) -> bool:
        """If `True` source data is a change data capture (CDC)"""
        return self.cdc is not None

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, data source is used in the context of a DLT pipeline"""
        is_orchestrator_dlt = False
        if self._parent and self._parent.is_orchestrator_dlt:
            is_orchestrator_dlt = True
        return is_orchestrator_dlt

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def read(self, spark=None) -> AnyDataFrame:
        """
        Read data with options specified in attributes.

        Parameters
        ----------
        spark:
            Spark context

        Returns
        -------
        : DataFrame
            Resulting dataframe
        """
        if self.mock_df is not None:
            df = self.mock_df
        elif self.dataframe_type == "SPARK":
            df = self._read_spark(spark=spark)
        elif self.dataframe_type == "POLARS":
            df = self._read_polars()
        else:
            raise ValueError(
                f"DataFrame type '{self.dataframe_type}' is not supported."
            )

        if is_spark_dataframe(df):
            df = self._post_read_spark(df)
        elif is_polars_dataframe(df):
            df = self._post_read_polars(df)

        return df

    def _read_spark(self, spark) -> SparkDataFrame:
        raise NotImplementedError()

    def _read_polars(self) -> PolarsDataFrame:
        raise NotImplementedError()

    def _post_read_spark(self, df: SparkDataFrame) -> SparkDataFrame:

        import pyspark.sql.functions as F

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops)

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Renames
        if self.renames:
            for old_name, new_name in self.renames.items():
                df = df.withColumnRenamed(old_name, new_name)

        # Apply Watermark
        if self.watermark:
            df = df.withWatermark(
                self.watermark.column,
                self.watermark.threshold,
            )

        # Broadcast
        if self.broadcast:
            df = F.broadcast(df)

        # Sample
        if self.sample:
            df = df.sample(fraction=self.sample.fraction, seed=self.sample.seed)

        # Limit
        if self.limit:
            df = df.limit(self.limit)

        return df

    def _post_read_polars(self, df: PolarsDataFrame) -> PolarsDataFrame:

        from laktory.polars.expressions.sql import _parse_token
        import polars as pl

        # Apply filter
        if self.filter:
            df = df.filter(pl.Expr.laktory.sql_expr(self.filter))

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [_parse_token(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [_parse_token(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops)

        # Renames
        if self.renames:
            df = df.rename(self.renames)

        # Apply Watermark
        if self.watermark:
            raise NotImplementedError(
                "Watermarking not supported with POLARS dataframe"
            )

        # Broadcast
        if self.broadcast:
            raise NotImplementedError(
                "Broadcasting not supported with POLARS dataframe"
            )

        # Sample
        if self.sample:
            df = df.sample(fraction=self.sample.fraction, seed=self.sample.seed)

        # Limit
        if self.limit:
            df = df.limit(self.limit)

        return df



import os
from typing import Literal
from typing import Union
from laktory.models.datasinks.basedatasink import BaseDataSink
from laktory.spark import SparkDataFrame
from laktory.polars import PolarsDataFrame
from laktory.models.datasources.filedatasource import FileDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class FileDataSink(BaseDataSink):
    """
    Disk file(s) data sink such as csv, parquet or Delta Table.

    Attributes
    ----------
    checkpoint_location:
        Path to which the checkpoint file for streaming dataframe should
        be written. If `None`, parent directory of `path` is used.
    format:
        Format of the data files
    path:
        Path to which the DataFrame needs to be written.

    Examples
    ---------
    ```python
    from laktory import models
    import pandas as pd

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    sink = models.FileDataSink(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="PARQUET",
        mode="OVERWRITE",
    )
    # sink.write(df)
    ```
    """

    checkpoint_location: Union[str, None] = None
    format: Literal["CSV", "PARQUET", "DELTA", "JSON", "EXCEL"] = "DELTA"
    path: str

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    @property
    def _checkpoint_location(self):
        if self.checkpoint_location:
            return self.checkpoint_location
        return os.path.dirname(self.path)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _write_spark(self, df: SparkDataFrame, mode=None) -> None:

        if self.format in ["EXCEL"]:
            raise ValueError(f"'{self.format}' format is not supported with Spark")

        # Default Options
        _options = {"mergeSchema": "true"}
        if self.mode in ["OVERWRITE", "COMPLETE"]:
            _options["mergeSchema"] = "false"
        if df.isStreaming:
            _options["checkpointLocation"] = self._checkpoint_location

        # User Options
        for k, v in self.write_options.items():
            _options[k] = v

        if df.isStreaming:
            logger.info(f"Writing df as stream {self.format} to {self.path}")

            writer = df.writeStream.trigger(
                availableNow=True
            )  # TODO: Add option for trigger?

        else:
            logger.info(f"Writing df as static {self.format} to {self.path}")
            writer = df.write

        writer = writer.mode(mode).format(self.format).options(**_options)

        writer.save(self.path)

    def _write_polars(self, df: PolarsDataFrame, mode=None) -> None:

        isStreaming = False

        if isStreaming:
            pass
            logger.info(f"Writing df as stream {self.format} to {self.path}")

        else:
            logger.info(f"Writing df as static {self.format} to {self.path}")

        if self.format != "DELTA":
            if mode:
                raise ValueError(
                    "'mode' configuration with Polars only supported by 'DELTA' format"
                )
        else:
            if not mode:
                raise ValueError(
                    "'mode' configuration required with Polars 'DELTA' format"
                )

        if self.format == "CSV":
            df.write_csv(self.path, **self.write_options)
        elif self.format == "DELTA":
            df.write_delta(self.path, mode=mode, **self.write_options)
        elif self.format == "EXCEL":
            df.write_excel(self.path, **self.write_options)
        elif self.format == "JSON":
            df.write_json(self.path, **self.write_options)
        elif self.format == "PARQUET":
            df.write_parquet(self.path, **self.write_options)

    def as_source(self, as_stream: bool = None) -> FileDataSource:
        """
        Generate a file data source with the same path as the sink.

        Parameters
        ----------
        as_stream:
            If `True`, sink will be read as stream.

        Returns
        -------
        :
            File Data Source
        """

        source = FileDataSource(
            path=self.path,
            format=self.format,
        )
        if as_stream:
            source.as_stream = as_stream

        return source
